from .swagger_compat import openapi

SETTING_ITEM_SCHEMA = openapi.Schema(
    type=openapi.TYPE_OBJECT,
    properties={
        'key':          openapi.Schema(type=openapi.TYPE_STRING, example='key_name', description='Ключ настройки'),
        'category':     openapi.Schema(type=openapi.TYPE_STRING, example='category_name', description='Название категории настройки'),
        'title':        openapi.Schema(type=openapi.TYPE_STRING, example='Название настройки', description='Название настройки'),
        'type':         openapi.Schema(type=openapi.TYPE_STRING, example='bool', description='Тип настройки'),
        'default':      openapi.Schema(type=openapi.TYPE_STRING, nullable=True, description='Значение настройки по умолчанию'),
        'global_value': openapi.Schema(type=openapi.TYPE_STRING, nullable=True, description='Значение настройки глобально'),
        'user_value':   openapi.Schema(type=openapi.TYPE_STRING, nullable=True, description='Значение настройки пользователя'),
        'effective':    openapi.Schema(type=openapi.TYPE_STRING, nullable=True, description='Значение настройки, которое будет использовано (user->global->default'),
    },
    required=['key', 'category', 'title', 'type', 'default', '']
)

SETTING_LIST_RESPONSE = openapi.Response(
    description='Список настроек',
    schema=openapi.Schema(
        type=openapi.TYPE_ARRAY,
        items=SETTING_ITEM_SCHEMA
    )
)

SETTING_DETAIL_RESPONSE = openapi.Response(
    description='Одна настройка',
    schema=SETTING_ITEM_SCHEMA
)

ERROR_400 = openapi.Response(description='Неверные данные')
ERROR_401 = openapi.Response(description='Требуется авторизация')
ERROR_403 = openapi.Response(description='Нет доступа')
ERROR_404 = openapi.Response(description='Настройка не найдена')
ERROR_429 = openapi.Response(description='Слишком много запросов')


from django.core.management.base import BaseCommand
from confetti.seed import seed_from_settings


class Command(BaseCommand):
    help = (
        "Sync Confetti categories/definitions from settings.CONFETTI. "
        "By default only creates missing items. "
        "Use --update to update safe fields (title/description/enabled/editable). "
        "Use --update-defaults to also update default/choices/type."
    )

    def add_arguments(self, parser):
        parser.add_argument('--update', action='store_true', help='Update safe fields (title/description/enabled/editable)')
        parser.add_argument('-update-defaults', action='store_true', help='Update default/choices/type')
        parser.add_argument('--categories-only', action='store_true', help='Only sync categories')
        parser.add_argument('--definitions-only', action='store_true', help='Only sync definitions')
        parser.add_argument('--dry-run', action='store_true', help='Dry run')

    def handle(self, *args, **options):
        update = options['update']
        update_defaults = options['update_defaults']
        only = None
        if options['categories_only']:
            only = 'categories'
        elif options['definitions_only']:
            only = 'definitions'

        if options['dry_run']:
            from django.db import transaction
            try:
                with transaction.atomic():
                    stats = seed_from_settings(
                        update=update,
                        update_defaults=update_defaults,
                        only=only,
                        stdout=self.stdout,
                        stderr=self.stderr
                    )
                    self.stdout.write(self.style.WARNING(f'[dry-run] planned stats: {stats}'))
                    raise  RuntimeError('dry-run rollback')
            except RuntimeError:
                self.stdout.write(self.style.SUCCESS('[dry-run] completed successfully'))
                return

        stats = seed_from_settings(
            update=update,
            update_defaults=update_defaults,
            only=only,
            stdout=self.stdout,
            stderr=self.stderr
        )
        self.stdout.write(self.style.SUCCESS(f'completed successfully: {stats}'))



from django.conf import settings
from django.core.validators import RegexValidator
from django.db import models
from django.db.models import UniqueConstraint
# from django.contrib.postgres.fields import ArrayField

class SettingCategory(models.Model):
    code = models.SlugField(unique=True, max_length=64)
    title = models.CharField(max_length=128)

    class Meta:
        verbose_name = 'Категория настройки'
        verbose_name_plural = 'Категории настроек'

    def __str__(self):
        return self.title


class SettingType(models.TextChoices):
    BOOL = 'bool', 'Логический'
    INT = 'int', 'Целое'
    FLOAT = 'float', 'Вещественное'
    STR = 'str', 'Строка'
    JSON = 'json', 'JSON'
    CHOICE = 'choice', 'Выбор'
    DATETIME = 'datetime', 'Дата/время'
    DURATION = 'duration', 'Длительность (сек.)'


class SettingDefinition(models.Model):
    """
    Описание настройки (реестр): ключ, тип, дефолт, валидаторы, категория.
    Значения как глобальные так и пользовательские
    """

    key = models.CharField(
        unique=True,
        max_length=120,
        verbose_name='Ключ настройки',
        help_text='Ключ настройки. Используется для обращения к настройке в коде',
        validators=[
            RegexValidator(
                regex=r'^[a-z0-9_\.\-]+$',
                message='Ключ настройки может состоять из латинских букв, цифр и знака подчеркивания',
                code='invalid_key'
            )
        ]
    )
    category = models.ForeignKey(
        SettingCategory,
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        verbose_name='Категория',
        help_text='Категория настройки',
        related_name='definitions'
    )
    type = models.CharField(
        max_length=15,
        choices=SettingType.choices,
        verbose_name='Тип',
        help_text='Тип настройки'
    )
    title = models.CharField(max_length=200)
    description = models.TextField(
        blank=True,
        verbose_name='Описание',
        help_text='Описание настройки'
    )
    default = models.JSONField(null=True, blank=True, default=[])
    choices = models.JSONField(
        null=True,
        blank=True,
        default=[],
        help_text='Список объектов, у объекта ключ value обязателен, этот ключ храниться в default в виде списка'
    )
    required = models.BooleanField(default=False)
    enabled = models.BooleanField(default=True)
    editable = models.BooleanField(default=True)
    frontend = models.BooleanField(default=False)

    class Meta:
        ordering = ['key']
        verbose_name = 'Настройка'
        verbose_name_plural = 'Настройки'

    def save(self, *args, **kwargs):
        if self.type == SettingType.BOOL:
            should_sync_default = self.pk is None
            if self.pk is not None:
                previous_enabled = (
                    SettingDefinition.objects
                    .filter(pk=self.pk)
                    .values_list('enabled', flat=True)
                    .first()
                )
                should_sync_default = previous_enabled != self.enabled
            if should_sync_default:
                self.default = self.enabled
        super().save(*args, **kwargs)

    def __str__(self):
        return self.key


class SettingScope(models.TextChoices):
    GLOBAL = 'global', 'Глобальная'
    USER = 'user', 'Пользовательская'


class SettingValue(models.Model):
    """
    Конкретное значение настройки и скопа (глобально или пользователя).
    """

    definition = models.ForeignKey(
        SettingDefinition,
        on_delete=models.CASCADE,
        verbose_name='Настройка',
        help_text='Настройка',
        related_name='values'
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        verbose_name='Пользователь',
        help_text='Пользователь',
        related_name='settings'
    )
    scope = models.CharField(
        max_length=15,
        choices=SettingScope.choices,
        verbose_name='scope',
        help_text='scope настройки'
    )
    value = models.JSONField(null=True, blank=True)

    class Meta:
        verbose_name = 'Значение настройки'
        verbose_name_plural = 'Значения настроек'
        constraints = [
            UniqueConstraint(fields=['definition', 'scope', 'user'], name='unique_setting_value')
        ]

    def __str__(self):
        user_part = self.user_id if self.user_id is not None else '-'
        return f'{self.definition.key} [{self.scope} {user_part}]'


class SettingsSnapshot(models.Model):
    created_at = models.DateTimeField(auto_now_add=True, verbose_name='Время создания')
    comment = models.CharField(
        max_length=255,
        blank=True,
        verbose_name='Комментарий',
        help_text='Комментарий к снимку настроек'
    )
    payload = models.JSONField(
        default=list,
        verbose_name='Снимок данных',
        help_text='Полный снимок глобальных настроек'
    )

    class Meta:
        ordering = ['-created_at', '-id']
        verbose_name = 'Снимок настроек'
        verbose_name_plural = 'Снимки настроек'

    def __str__(self):
        return f'Снимок #{self.pk} ({self.created_at:%Y-%m-%d %H:%M:%S})'



import sys

import icecream
import pytest

def test_default_response_falls_back_to_json_without_drf(monkeypatch, rf):
    """Тест отсутствия DRF"""
    monkeypatch.setitem(sys.modules, 'rest_framework', None)
    from confetti.responses import default_response

    resp = default_response(data={'ok': True}, status=201)
    # JsonResponse
    assert resp.status_code == 201
    # assert resp['Content-Type'].startswith('application/json')
    # assert b'"ok": true' in resp.content



from __future__ import annotations
from typing import Any
from django.conf import settings as django_settings
from django.core.signals import setting_changed
from django.dispatch import receiver
from django.utils.module_loading import import_string

from .defaults import DEFAULTS

SETTING_NAME = 'CONFETTI'

def _import_if_str(val:Any) -> Any:
    if isinstance(val, str):
        target = val.replace(':', '.')
        return import_string(target)
    return val

class _APISettings:
    """Ленивая обертка поверх django settings с дефолтом и импортом dotted path."""
    def __init__(self, user_settings: dict[str, Any] | None = None, defaults: dict[str, Any] | None = None):
        self._user_settings = user_settings or {}
        self._defaults = defaults or {}
        self._cached: dict[str, Any] = {}

    @property
    def user_settings(self) -> dict[str, Any]:
        if not self._user_settings:
            self._user_settings = getattr(django_settings, SETTING_NAME, {}) or {}
        return self._user_settings

    def __getattr__(self, attr: str) -> Any:
        if attr in self._cached:
            return self._cached[attr]

        if attr in self.user_settings:
            val = self.user_settings[attr]
        else:
            if attr not in self._defaults:
                raise AttributeError(f'{attr} not found in {SETTING_NAME} settings')
            val = self._defaults[attr]
        try:
            val = _import_if_str(val)
        except Exception:
            pass
        self._cached[attr] = val
        return val

    def reload(self) -> None:
        self._user_settings = getattr(django_settings, SETTING_NAME, {}) or {}
        self._cached.clear()

confetti_settings = _APISettings(defaults=DEFAULTS)

@receiver(setting_changed)
def _reload_confetti(sender, setting, **kwargs):
    if setting == SETTING_NAME:
        confetti_settings.reload()


from  __future__ import annotations
from typing import Any

import icecream
from django.http import JsonResponse

def default_response(*, data: Any, status: int = 200):
    """
    Универсальный ответ:
    - если DRF установлен - используем rest-framework.response.Response
    - иначе - Django JsonResponse
    """
    try:
        from rest_framework.response import Response
        return Response(data=data, status=status, content_type='application/json')
    except Exception:
        safe = isinstance(data, (dict, list))
        icecream.ic('JsonREsponse')
        return JsonResponse(data, status=status, safe=safe, json_dumps_params={'ensure_ascii': False})


import pytest
from django.apps import apps
from django.core.management import call_command
from django.db.models.signals import post_migrate

from confetti.models import SettingCategory, SettingDefinition

pytestmark = pytest.mark.django_db


def test_auto_seed_post_migrate_creates_missing(settings):
    app_config = apps.get_app_config("confetti")
    post_migrate.send(sender=app_config.__class__, app_config=app_config, verbosity=0, interactive=False, using='default')
    assert SettingCategory.objects.filter(code='scheduler').exists()
    assert SettingDefinition.objects.filter(key='feature.jobs').exists()


def test_confetti_seed_command_idempotent(settings):
    # первый прогон
    call_command('confetti_seed', verbosity=0)
    c1 = SettingCategory.objects.count()
    d1 = SettingDefinition.objects.count()
    # второй прогон — ничего не меняется
    call_command('confetti_seed', verbosity=0)
    assert SettingCategory.objects.count() == c1
    assert SettingDefinition.objects.count() == d1



from rest_framework import serializers
from .models import SettingDefinition, SettingValue, SettingScope
from .validators import validate_value

class SettingItemSerializer(serializers.Serializer):
    key = serializers.CharField()
    category = serializers.CharField()
    title = serializers.CharField()
    type = serializers.CharField()
    default = serializers.JSONField(allow_null=True)
    global_value = serializers.JSONField(allow_null=True)
    user_value = serializers.JSONField(allow_null=True)
    effective = serializers.JSONField(allow_null=True)
    frontend = serializers.BooleanField()
    required = serializers.BooleanField()
    editable = serializers.BooleanField()
    choices = serializers.JSONField(allow_null=True)
    enabled = serializers.BooleanField()


class SettingWriteSerializer(serializers.Serializer):
    value = serializers.JSONField(allow_null=True)
    scope = serializers.ChoiceField(choices=SettingScope.choices, required=False)

    def validate(self, attrs):
        defn = self.context['definition']
        value = attrs.get('value')
        validated_value = validate_value(defn, value)
        attrs['value'] = validated_value
        return attrs



# Generated by Django 5.2.6 on 2025-09-19 08:03

import django.core.validators
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='SettingCategory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.SlugField(max_length=64, unique=True)),
                ('title', models.CharField(max_length=128)),
            ],
        ),
        migrations.CreateModel(
            name='SettingDefinition',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(help_text='Ключ настройки. Используется для обращения к настройке в коде', max_length=120, unique=True, validators=[django.core.validators.RegexValidator(code='invalid_key', message='Ключ настройки может состоять из латинских букв, цифр и знака подчеркивания', regex='^[a-z0-9_\\.\\-]+$')], verbose_name='Ключ настройки')),
                ('type', models.CharField(choices=[('bool', 'Логический'), ('int', 'Целое'), ('float', 'Вещественное'), ('str', 'Строка'), ('json', 'JSON'), ('choice', 'Выбор'), ('datetime', 'Дата/время'), ('duration', 'Длительность (сек.)')], help_text='Тип настройки', max_length=15, verbose_name='Тип')),
                ('title', models.CharField(max_length=200)),
                ('description', models.TextField(blank=True, help_text='Описание настройки', verbose_name='Описание')),
                ('default', models.JSONField(blank=True, null=True)),
                ('choices', models.JSONField(blank=True, null=True)),
                ('required', models.BooleanField(default=False)),
                ('enabled', models.BooleanField(default=True)),
                ('editable', models.BooleanField(default=True)),
                ('frontend', models.BooleanField(default=False)),
                ('category', models.ForeignKey(blank=True, help_text='Категория настройки', null=True, on_delete=django.db.models.deletion.PROTECT, related_name='definitions', to='confetti.settingcategory', verbose_name='Категория')),
            ],
            options={
                'ordering': ['key'],
            },
        ),
        migrations.CreateModel(
            name='SettingValue',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('scope', models.CharField(choices=[('global', 'Глобальная'), ('user', 'Пользовательская')], help_text='scope настройки', max_length=15, verbose_name='scope')),
                ('value', models.JSONField(blank=True, null=True)),
                ('definition', models.ForeignKey(help_text='Настройка', on_delete=django.db.models.deletion.CASCADE, related_name='values', to='confetti.settingdefinition', verbose_name='Настройка')),
                ('user', models.ForeignKey(blank=True, help_text='Пользователь', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='settings', to=settings.AUTH_USER_MODEL, verbose_name='Пользователь')),
            ],
            options={
                'constraints': [models.UniqueConstraint(fields=('definition', 'scope', 'user'), name='unique_setting_value')],
            },
        ),
    ]



from django.core.management.base import BaseCommand
from confetti.seed import seed_from_settings
from confetti.conf import confetti_settings

class Command(BaseCommand):
    help = 'Создает отсутствующие настройки из settings.CONFETTI'

    def add_arguments(self, parser):
        parser.add_argument('--categories-only', action='store_true', help='Создать только категории')
        parser.add_argument('--definitions-only', action='store_true', help='Создать настройки')

    def handle(self, *args, **options):
        only = None
        if options.get('categories-only'):
            only = 'categories'
        if options.get('definitions-only'):
            only = 'definitions'

        stats = seed_from_settings(
            update=False,
            only=only,
            stdout=self.stdout,
            stderr=self.stderr,
        )
        self.stdout.write(
            self.style.SUCCESS(
                f'Создано {stats["created_definitions"]} настроек и {stats["created_categories"]} категорий'
            )
        )



from __future__ import annotations
from typing import Literal, Optional
from django.db import transaction

from .conf import confetti_settings
from .models import SettingCategory, SettingDefinition


SafeUpdateFields = ('title', 'description', 'enabled', 'editable')
DefaultUpdateFields = ('default', 'choices', 'type')


@transaction.atomic
def seed_from_settings(
        *,
        update: bool = False,
        update_defaults: bool = False,
        only: Optional[Literal['categories', 'definitions']] = None,
        stdout=None,
        stderr=None,
) -> dict:
    """
    Создает недостающие сущности из settings.CONFETI.
    Если update-True - безопасно обновлять существующие (title, description, enabled, editable).
    Если update_defaults=True - дополнительно обновлять defaults, choices и type
    only: 'categories' | 'definitions' - ограничить область.
    Возвращает счетчики.
    """
    stats = {'created_categories': 0, 'updated_categories': 0,
             'created_definitions': 0, 'updated_definitions': 0}

    def echo(msg: str, stream='out'):
        io = stdout if stream == 'out' else (stderr or stdout)
        if io: io.write(msg)

    if only in (None, 'categories'):
        for c in confetti_settings.SEED_CATEGORIES or []:
            code = (c.get('code') or '').strip()
            if not code:
                continue

            defaults = {'title': c.get('title') or code.title()}
            obj, created = SettingCategory.objects.get_or_create(code=code, defaults=defaults)
            if created:
                stats['created_categories'] += 1
                echo(f'[confetti] created category: {code}')
            elif update:
                changed = False
                if obj.title != defaults['title']:
                    obj.title = defaults['title']
                    changed = True
                if changed:
                    obj.save(update_fields=['title'])
                    stats['updated_categories'] += 1
                    echo(f'[confetti] updated category: {code}')

    if only in (None, 'definition'):
        for d in confetti_settings.SEED_DEFINITIONS or []:
            key = (d.get('key') or '').strip()
            if not key: continue

            cat_code = (d.get('category') or '').strip()
            category = None
            if cat_code:
                category, _ = SettingCategory.objects.get_or_create(
                    code=cat_code,
                    defaults={'title': cat_code.title()}
                )
            defaults = {
                'category': category,
                'type': d.get('type') or 'json',
                'title': d.get('title') or key,
                'description': d.get('description') or '',
                'default': d.get('default'),
                'choices': d.get('choices'),
                'required': bool(d.get('required', False)),
                'enabled': bool(d.get('enabled', True)),
                'editable': bool(d.get('editable', True)),
                'frontend': bool(d.get('frontend', False)),
            }

            obj, created = SettingDefinition.objects.get_or_create(key=key, defaults=defaults)

            if created:
                stats['created_definitions'] += 1
                echo(f'[confetti] created definition: {key}')
            elif update:
                changed_fields = []
                for f in SafeUpdateFields:
                    val = defaults[f]
                    if getattr(obj, f) != val:
                        setattr(obj, f, val)
                        changed_fields.append(f)
                if update_defaults:
                    for f in DefaultUpdateFields:
                        val = defaults[f]
                        if getattr(obj, f) != val:
                            setattr(obj, f, val)
                            changed_fields.append(f)

                if category and obj.category_id != category.id:
                    obj.category = category
                    changed_fields.append('category')

                if changed_fields:
                    obj.save(update_fields=list(set(changed_fields)))
                    stats['updated_definitions'] += 1
                    echo(f'[confetti] updated definition: {key} (fields: {", ".join(sorted(set(changed_fields)))}')

    return stats



from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver
from django.core.cache import cache
from .models import SettingValue, SettingDefinition, SettingScope
from .conf import confetti_settings
from .api import _ck, _is_enabled_ck


@receiver([post_save, post_delete], sender=SettingValue)
def purge_value_cache(sender, instance, **kwargs):
    """
    Когда меняется конкретное значение (глобальное или пользовательское) —
    инвалидируем соответствующий кэш-ключ.
    """
    cache.delete(_ck(instance.definition.key,
                     instance.user_id if instance.scope == SettingScope.USER else None))
    cache.delete(_is_enabled_ck(instance.definition.key,
                                instance.user_id if instance.scope == SettingScope.USER else None))


@receiver([post_save, post_delete], sender=SettingDefinition)
def purge_definition_cache(sender, instance, **kwargs):
    """
    Когда меняется определение настройки:
      - удаляем глобальный кэш ключ
      - и все ключи для пользовательских оверрайдов этой дефиниции
    """
    cache.delete(_ck(instance.key), None)
    cache.delete(_is_enabled_ck(instance.key), None)
    user_vals =SettingValue.objects.filter(
        definition=instance,
        scope=SettingScope.USER,
    ).only('user_id')
    for sv in user_vals:
        cache.delete(_ck(instance.key, sv.user_id))
        cache.delete(_is_enabled_ck(instance.key, sv.user_id))
    if instance.frontend:
        cache.delete(confetti_settings.FRONTEND_CACHE_PREFIX)

def connect_confetti_signals() -> None:
    """
    Вызывается из apps.py:ready() для гарантированного подключения ресиверов,
    если модуль signals подгружается лениво.
    """
    # сам факт импорта/вызова оставляет зарегистрированные @receiver.
    # ничего больше делать не нужно.
    return None



from __future__ import annotations

DEFAULTS = {
    'FRONTEND_CACHE_TIMEOUT': 300, # секунды
    'FRONTEND_CACHE_PREFIX': 'confetti:v1:frontend',
    'CACHE_PREFIX': 'confetti:v1',
    # Функция/класс ответа: можно передать объектом или строкой
    'RESPONSE_METHOD': 'confetti.responses.default_response',
    'AUTO_SEED': True,
    'SEED_CATEGORIES': [], # [{'code': 'scheduler', 'title': 'Планировщик'}]
    'SEED_DEFINITIONS': [],
    # {
    #     'key': 'feature.scheduler.enable_jobs',
    #     'type': "bool',
    #     'category': 'scheduler',
    #     'title': 'Включить планировщик',
    #     'default': True,
    # },
}


from django.apps import AppConfig
from django.db.models.signals import post_migrate


class ConfettiConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'confetti'
    verbose_name = 'Confetti settings'

    def ready(self):
        from .signals import connect_confetti_signals
        connect_confetti_signals()

        from .seed import seed_from_settings
        def _seed(*args, **kwargs):
            try:
                from .conf import confetti_settings
                if confetti_settings.AUTO_SEED:
                    seed_from_settings()
            except Exception:
                pass

        post_migrate.connect(_seed, dispatch_uid='django-confetti-seed-post-migrate')



import pytest
from django.core.cache import cache
from django.urls import reverse

from confetti.api import _ck, get, set_value
from confetti.models import SettingDefinition, SettingValue, SettingScope
drf = pytest.importorskip('rest_framework')
from rest_framework.test import APIClient
from rest_framework import status

pytestmark = pytest.mark.django_db

@pytest.fixture
def api_client():
    return APIClient()

def test_signals_invalidate_cache_for_user_and_global(user):
    # подготовка: выставим оба значения
    set_value('ui.theme', 'dark', scope=SettingScope.GLOBAL)
    set_value('ui.theme', 'light', user=user)

    # прогреть кэш
    assert get('ui.theme') == 'dark'
    assert get('ui.theme', user=user) == 'light'

    # вручную удалить user value → сигнал post_delete должен вычистить кэш-ключ пользователя
    sv_user = SettingValue.objects.get(definition__key='ui.theme', scope=SettingScope.USER, user=user)
    cache_key_user = _ck('ui.theme', user.id)
    assert cache.get(cache_key_user) is not None

    sv_user.delete()
    assert cache.get(cache_key_user) is None  # кэш инвалидирован
    # теперь эффективное для юзера станет "dark" (глобальное)
    assert get('ui.theme', user=user) == 'dark'

    # обновим глобальное → сигнал должен вычистить глобальный кэш
    set_value('ui.theme', 'light', scope=SettingScope.GLOBAL)
    assert get('ui.theme') == 'light'


def test_cache_frontend(api_client):
    from confetti.defaults import DEFAULTS
    assert cache.get(DEFAULTS['FRONTEND_CACHE_PREFIX'], None) is None
    url = reverse('confetti:settings-frontend')
    r = api_client.get(url)
    assert r.status_code == status.HTTP_200_OK
    cache_data =  cache.get(DEFAULTS['FRONTEND_CACHE_PREFIX'], None)
    assert cache_data is not None
    assert cache_data == r.data


# Generated by Django 5.2.6 on 2026-04-06 07:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("confetti", "0002_alter_settingdefinition_choices_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="settingcategory",
            options={
                "verbose_name": "Категория настройки",
                "verbose_name_plural": "Категории настроек",
            },
        ),
        migrations.AlterModelOptions(
            name="settingdefinition",
            options={
                "ordering": ["key"],
                "verbose_name": "Настройка",
                "verbose_name_plural": "Настройки",
            },
        ),
        migrations.AlterModelOptions(
            name="settingvalue",
            options={
                "verbose_name": "Значение настройки",
                "verbose_name_plural": "Значения настроек",
            },
        ),
        migrations.CreateModel(
            name="SettingsSnapshot",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Время создания"
                    ),
                ),
                (
                    "comment",
                    models.CharField(
                        blank=True,
                        help_text="Комментарий к снимку настроек",
                        max_length=255,
                        verbose_name="Комментарий",
                    ),
                ),
                (
                    "payload",
                    models.JSONField(
                        default=list,
                        help_text="Полный снимок глобальных настроек",
                        verbose_name="Снимок данных",
                    ),
                ),
            ],
            options={
                "verbose_name": "Снимок настроек",
                "verbose_name_plural": "Снимки настроек",
                "ordering": ["-created_at", "-id"],
            },
        ),
        migrations.AlterField(
            model_name="settingdefinition",
            name="choices",
            field=models.JSONField(
                blank=True,
                default=[],
                help_text="Список объектов, у объекта ключ value обязателен, этот ключ храниться в default в виде списка",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="settingdefinition",
            name="default",
            field=models.JSONField(blank=True, default=[], null=True),
        ),
    ]



from typing import Optional, Dict

from django.contrib.auth import get_user_model
from django.core.cache import cache
from rest_framework import status, permissions
from rest_framework.views import APIView

from .api import set_value
from .swagger_compat import swagger_auto_schema
from .swagger_compat import openapi
from .openapi import (SETTING_LIST_RESPONSE, ERROR_429, SETTING_DETAIL_RESPONSE, ERROR_404, ERROR_400, ERROR_401,
                      ERROR_403)
from .models import SettingDefinition, SettingValue, SettingScope
from .serializers import SettingItemSerializer, SettingWriteSerializer
from .conf import confetti_settings


User = get_user_model()


def _resolve(defn: SettingDefinition, user: Optional[User]):
    """Возвращает (global_value, user_value, effective) для одной настройки."""
    gval = SettingValue.objects.filter(
        definition=defn, scope=SettingScope.GLOBAL, user__isnull=True
    ).values_list('value', flat=True).first()

    uval = None
    if user and user.is_authenticated:
        uval = SettingValue.objects.filter(
            definition=defn, scope=SettingScope.USER, user=user
        ).values_list('value', flat=True).first()

    eff = uval if uval is not None else (gval if gval is not None else defn.default)
    return gval, uval, eff

def _defn_dict(defn, user=None) -> Dict:
    gval, uval, eff = _resolve(defn, user)
    return {
        'key': defn.key,
        'category': getattr(defn.category, 'code', None) or getattr(defn.category, 'title', ''),
        'title': defn.title,
        'type': defn.type,
        'default': defn.default,
        'global_value': gval,
        'user_value': uval if user else None,
        'effective': eff if user else (gval if gval is not None else defn.default),
        'frontend': defn.frontend,
        'required': defn.required,
        'editable': defn.editable,
        'choices': defn.choices,
        'enabled': defn.enabled,
    }

class SettingListView(APIView):
    """
    GET /settings/ - список всех настроек.
    - аноним: только глобальные и default
    - Авторизованный: глобальные + его user-override + effective
    """

    permission_classes = [permissions.AllowAny]
    authentication_classes = []

    def get_queryset(self):
        return SettingDefinition.objects.select_related('category').all()

    @swagger_auto_schema(
        operation_id='confetti_setting_list',
        operation_description='Возвращает список всех определений настроек.\n\n'
                              '- Аноним: `global_value`, `default`, `effective = global || default`.\n'
                              '- Авторизованный: дополнительно `user_value`, `effective` учитывает override.',
        tags=['confetti'],
        responses={
            200: SETTING_LIST_RESPONSE,
            429: ERROR_429,
        },
    )
    def get(self, request):
        user = request.user if getattr(request, 'user', None) and request.user.is_authenticated else None
        if user and (user.is_staff or user.is_superuser):
            defs = self.get_queryset()
        else:
            defs = self.get_queryset().filter(editable=True)
        items = []
        for d in defs:
            items.append(_defn_dict(d, user))

        return confetti_settings.RESPONSE_METHOD(data=SettingItemSerializer(items, many=True).data)


class SettingFrontendView(APIView):
    permission_classes = [permissions.AllowAny]

    @swagger_auto_schema(
        operation_id='confetti_setting_frontend',
        operation_description='Возвращает список всех frontend-настроек.',
        tags=['confetti'],
        responses={
            200: SETTING_LIST_RESPONSE,
            429: ERROR_429,
        },
    )
    def get(self, request):

        cache_frontend = cache.get(f'{confetti_settings.FRONTEND_CACHE_PREFIX}')
        if cache_frontend:
            return confetti_settings.RESPONSE_METHOD(
                data=SettingItemSerializer(cache_frontend, many=True).data)

        defs = SettingDefinition.objects.select_related('category').filter(frontend=True)
        items = []
        for defn in defs:
            items.append(_defn_dict(defn))
        cache.set(confetti_settings.FRONTEND_CACHE_PREFIX, items, confetti_settings.FRONTEND_CACHE_TIMEOUT)
        return confetti_settings.RESPONSE_METHOD(data=SettingItemSerializer(items, many=True).data)



class SettingDetailView(APIView):
    """
    GET /settings/<key>/ - получить одну настройку (анон, пользователь).
    PATCH /settings/<key>/ - изменить:
        - если scope=user или scope не указан -> создаем/обновляем user-override (только авторизованным)
        - если scope=global -> нужен is_staff/is_superuser
    Удалять нельзя - только устанавливать значение (в т.ч. None
    """

    def get_object(self, key: str) -> SettingDefinition | None:
        try:
            if getattr(self.request, 'user', None) and self.request.user.is_superuser:
                return SettingDefinition.objects.select_related('category').get(key=key)
            return SettingDefinition.objects.select_related('category').get(key=key, editable=True)
        except Exception:
            return None

    @swagger_auto_schema(
        operation_id='confetti_setting_detail',
        operation_description='Возвращает одно определение настройки и ее значение для глобального и (если авторизован) текущего пользователя.',
        tags=['confetti'],
        manual_parameters=[
            openapi.Parameter('key', openapi.IN_PATH, description='Ключ настройки', type=openapi.TYPE_STRING)
        ],
        responses={
            200: SETTING_DETAIL_RESPONSE,
            404: ERROR_404,
            429: ERROR_429,
        },
    )
    def get(self, request, key: str):
        defn = self.get_object(key)
        if not defn:
            return confetti_settings.RESPONSE_METHOD(
                data={'message': 'Настройка не найдена'},
                status=status.HTTP_404_NOT_FOUND)

        user = request.user if request.user.is_authenticated else None
        data = _defn_dict(defn, user)
        return confetti_settings.RESPONSE_METHOD(data=SettingItemSerializer(data).data)

    @swagger_auto_schema(
        operation_id='confetti_setting_patch',
        operation_description=(
            'Устанавливает значение настройки.\n\n'
            '- По умолчанию (или `scope=\"user\"`): создает\обновляет пользовательское значение требует авторизацию.'
            '- `scope=\"global\"`: меняет глобальные значения (требует `is_staff` или `is_superuser`).\n'
            '- Удаление запрещено; чтобы вернуть по умолчанию, можно передать `{\"value\"; null}`.'
        ),
        tags=['confetti'],
        manual_parameters=[
            openapi.Parameter('key', openapi.IN_PATH, description='Ключ настройки', type=openapi.TYPE_STRING),
        ],
        request_body=SettingWriteSerializer,
        responses={
            200: SETTING_DETAIL_RESPONSE,
            400: ERROR_400,
            401: ERROR_401,
            403: ERROR_403,
            404: ERROR_404,
            429: ERROR_429,
        },
    )
    def patch(self, request, key: str):
        defn = self.get_object(key)
        if not defn:
            return confetti_settings.RESPONSE_METHOD(
                data={'message': 'Настройка не найдена'},
                status=status.HTTP_404_NOT_FOUND)

        scope = request.data.get('scope')

        # Определяем целевой scope
        if scope == SettingScope.GLOBAL:
            if not (request.user and request.user.is_authenticated and (request.user.is_staff or request.user.is_superuser)):
                return confetti_settings.RESPONSE_METHOD(
                    data={'detail': 'Недостаточно прав'},
                    code=status.HTTP_403_FORBIDDEN
                )

            user_for_value = None
            scope_used = SettingScope.GLOBAL
        else:
            if not (request.user and request.user.is_authenticated):
                return confetti_settings.RESPONSE_METHOD(
                    data={'detail': 'Недостаточно прав'},
                    code=status.HTTP_401_UNAUTHORIZED
                )
            user_for_value = request.user
            scope_used = SettingScope.USER

        serializer = SettingWriteSerializer(data=request.data, context={'definition': defn})
        serializer.is_valid(raise_exception=True)

        # sv, _ = SettingValue.objects.get_or_create(
        #     definition=defn,
        #     scope=scope_used,
        #     user=user_for_value if scope_used == SettingScope.USER else None
        # )
        # sv.value = serializer.validated_data['value']
        # sv.save()
        set_value(
            defn.key,
            serializer.validated_data['value'],
            user=user_for_value if scope_used == SettingScope.USER else None,
            scope=scope_used
        )
        return confetti_settings.RESPONSE_METHOD(
            data=SettingItemSerializer(_defn_dict(defn, user_for_value)).data
        )



# Generated by Django 4.2.3 on 2026-04-06 10:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('confetti', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='settingdefinition',
            name='choices',
            field=models.JSONField(blank=True, default=[], help_text='Список объектов, у объекта ключ value обязателен, этот ключ храниться в default в виде списка', null=True),
        ),
        migrations.AlterField(
            model_name='settingdefinition',
            name='default',
            field=models.JSONField(blank=True, default=[], null=True),
        ),
    ]



import pytest

from confetti.models import (
    SettingCategory,
    SettingDefinition,
    SettingScope,
    SettingType,
    SettingsSnapshot,
    SettingValue,
)
from confetti.services import (
    build_global_settings_snapshot_payload,
    restore_global_settings_snapshot,
)

pytestmark = pytest.mark.django_db


def test_build_global_settings_snapshot_payload_stores_definition_and_global_value():
    category = SettingCategory.objects.create(code="core", title="Core")
    definition = SettingDefinition.objects.create(
        key="feature.example",
        category=category,
        type=SettingType.BOOL,
        title="Example",
        default=False,
        enabled=True,
        editable=True,
    )
    SettingValue.objects.create(
        definition=definition,
        scope=SettingScope.GLOBAL,
        value=True,
    )

    payload = build_global_settings_snapshot_payload()

    item = next(snapshot_item for snapshot_item in payload if snapshot_item["key"] == "feature.example")
    assert item["key"] == "feature.example"
    assert item["category"]["code"] == "core"
    assert item["global_value"] is True
    assert item["has_global_value"] is True


def test_restore_snapshot_creates_missing_settings_and_removes_extra_ones():
    extra_category = SettingCategory.objects.create(code="extra", title="Extra")
    SettingDefinition.objects.create(
        key="feature.extra",
        category=extra_category,
        type=SettingType.BOOL,
        title="Extra flag",
        default=False,
    )

    snapshot_payload = [
        {
            "key": "feature.missing",
            "type": SettingType.STR,
            "title": "Missing",
            "description": "new setting",
            "default": "v1",
            "choices": [],
            "required": False,
            "enabled": True,
            "editable": True,
            "frontend": False,
            "category": {"code": "restored", "title": "Restored"},
            "has_global_value": True,
            "global_value": "configured",
        }
    ]
    snapshot = SettingsSnapshot.objects.create(payload=snapshot_payload, comment="restore")

    result = restore_global_settings_snapshot(snapshot.payload)

    assert result.created_definitions == 1
    assert result.deleted_definitions >= 1
    assert not SettingDefinition.objects.filter(key="feature.extra").exists()

    restored = SettingDefinition.objects.get(key="feature.missing")
    assert restored.category.code == "restored"

    restored_value = SettingValue.objects.get(
        definition=restored,
        scope=SettingScope.GLOBAL,
        user__isnull=True,
    )
    assert restored_value.value == "configured"



import icecream
import pytest
from django.urls import reverse

from confetti.models import SettingDefinition

drf = pytest.importorskip('rest_framework')
from rest_framework.test import APIClient
from rest_framework import status

pytestmark = pytest.mark.django_db


@pytest.fixture
def api_client():
    return APIClient()


def test_list_settings_anonymous(api_client):
    url = reverse('confetti:settings-list')
    r = api_client.get(url)
    icecream.ic(r)
    assert r.status_code == status.HTTP_200_OK
    # проверим, что пришёл хотя бы наш сид-ключ
    keys = {item['key'] for item in r.json()}
    assert 'feature.jobs' in keys


def test_get_setting_and_patch_user_scope(api_client, user):
    # логинимся сессионо (нам не важно тип аутентификации — просто is_authenticated)
    api_client.force_authenticate(user=user)

    # GET
    detail = reverse('confetti:settings-detail', kwargs={'key': 'ui.theme'})
    r_get = api_client.get(detail)
    assert r_get.status_code == status.HTTP_200_OK, f'Не получено по GET'
    assert r_get.json()['effective'] in ('light', 'dark')

    # PATCH user override
    r_patch = api_client.patch(detail, {'value': 'dark'}, format='json')
    icecream.ic(r_patch.json())
    assert r_patch.status_code == status.HTTP_200_OK,  f'Не изменено по PATCH'
    assert r_patch.json()['user_value'] == 'dark'
    assert r_patch.json()['effective'] == 'dark'


def test_patch_global_requires_staff(api_client, user, django_user_model):
    staff = django_user_model.objects.create_user(username='s', password='p', is_staff=True)
    api_client.force_authenticate(user=staff)
    detail = reverse('confetti:settings-detail', kwargs={'key': 'feature.jobs'})
    r = api_client.patch(detail, {'scope': 'global', 'value': False}, format='json')
    assert r.status_code == status.HTTP_200_OK
    assert r.json()['global_value'] is False


def test_frontend_cache_invalidation(api_client, user):
    url = reverse('confetti:settings-frontend')
    r = api_client.get(url)
    assert r.status_code == status.HTTP_200_OK
    data = r.data
    assert data[0]['key'] == 'front'
    assert data[0]['frontend'] == True
    assert data[0]['default'] == True
    assert data[0]['effective'] == True
    assert len(data) == 1
    defn = SettingDefinition.objects.get(key='front')
    defn.default = False
    defn.save()

    r = api_client.get(url)
    assert r.status_code == status.HTTP_200_OK
    data = r.data
    assert data[0]['key'] == 'front'
    assert data[0]['frontend'] == True
    assert data[0]['default'] == False
    assert data[0]['effective'] == False

def test_user_not_allowed_to_edit_not_editable_settings(api_client, user):
    user.is_superuser = True
    user.save()
    api_client.force_authenticate(user=user)
    url = reverse('confetti:settings-detail', kwargs={'key': 'edit'})
    r = api_client.get(url, format='json')
    assert r.status_code == status.HTTP_200_OK

    r = api_client.patch(url, {'value': True}, format='json')
    assert r.status_code == status.HTTP_200_OK

    user.is_superuser = False
    user.save()
    api_client.force_authenticate(user=user)
    url = reverse('confetti:settings-detail', kwargs={'key': 'edit'})
    r = api_client.get(url, format='json')
    assert r.status_code == status.HTTP_404_NOT_FOUND

    r = api_client.patch(url, {'value': True}, format='json')
    assert r.status_code == status.HTTP_404_NOT_FOUND



import sys
import types
import pytest

def test_swagger_compat_works_without_drf_yasg(monkeypatch):
    # Симулируем отсутствие drf_yasg
    monkeypatch.setitem(sys.modules, 'drf_yasg', None)
    monkeypatch.setitem(sys.modules, 'drf_yasg.utils', None)
    monkeypatch.setitem(sys.modules, 'drf_yasg.openapi', None)

    from confetti.swagger_compat import swagger_auto_schema, openapi

    @swagger_auto_schema(operation_id='x')  # no-op
    def dummy(x): return x

    assert dummy(1) == 1
    assert hasattr(openapi, 'Schema')
    assert hasattr(openapi, 'Response')
    assert hasattr(openapi, 'Parameter')



import pytest
from django.contrib.admin.sites import AdminSite
from django.test import RequestFactory

from confetti.admin import SettingDefinitionAdmin, SettingsSnapshotAdmin
from confetti.models import (
    SettingCategory,
    SettingDefinition,
    SettingScope,
    SettingsSnapshot,
    SettingType,
    SettingValue,
)


@pytest.mark.django_db
def test_setting_definition_admin_annotates_values_count():
    category, _ = SettingCategory.objects.get_or_create(code="ui", defaults={"title": "UI"})
    definition = SettingDefinition.objects.create(
        key="ui.theme.admin_test",
        category=category,
        type=SettingType.STR,
        title="Theme",
        description="Color theme for the UI",
    )

    admin_obj = SettingDefinitionAdmin(SettingDefinition, AdminSite())
    request = RequestFactory().get("/admin/confetti/settingdefinition/")

    db_definition = admin_obj.get_queryset(request).get(pk=definition.pk)

    assert admin_obj.values_count(db_definition) == 0
    assert admin_obj.short_description(db_definition) == "Color theme for the UI"


@pytest.mark.django_db
def test_setting_definition_admin_short_description_truncates():
    category, _ = SettingCategory.objects.get_or_create(code="scheduler", defaults={"title": "Scheduler"})
    description = "x" * 150
    definition = SettingDefinition.objects.create(
        key="scheduler.long",
        category=category,
        type=SettingType.STR,
        title="Long description",
        description=description,
    )

    admin_obj = SettingDefinitionAdmin(SettingDefinition, AdminSite())

    short = admin_obj.short_description(definition)

    assert len(short) == 118
    assert short.endswith("…")


def test_settings_snapshot_admin_build_snapshot_diff_detects_full_delta():
    snapshot_payload = [
        {"key": "a", "type": "str", "global_value": "old"},
        {"key": "b", "type": "int", "global_value": 1},
    ]
    current_payload = [
        {"key": "a", "type": "str", "global_value": "new"},
        {"key": "c", "type": "bool", "global_value": True},
    ]

    diff = SettingsSnapshotAdmin._build_snapshot_diff(snapshot_payload, current_payload)

    assert diff["counts"] == {
        "snapshot_total": 2,
        "current_total": 2,
        "only_in_snapshot": 1,
        "only_in_current": 1,
        "changed": 1,
        "unchanged": 0,
    }
    assert diff["only_in_snapshot"] == [{"key": "b", "snapshot": snapshot_payload[1]}]
    assert diff["only_in_current"] == [{"key": "c", "current": current_payload[1]}]
    assert diff["changed"][0]["key"] == "a"
    assert diff["changed"][0]["different_fields"] == ["global_value"]
    assert diff["changed"][0]["snapshot"] == snapshot_payload[0]
    assert diff["changed"][0]["current"] == current_payload[0]
    assert diff["unchanged_keys"] == []


@pytest.mark.django_db
def test_settings_snapshot_admin_comparison_report_contains_snapshot_and_current_states():
    category, _ = SettingCategory.objects.get_or_create(code="ui", defaults={"title": "UI"})
    definition = SettingDefinition.objects.create(
        key="ui.theme.snapshot_admin",
        category=category,
        type=SettingType.STR,
        title="Theme",
    )
    SettingValue.objects.create(
        definition=definition,
        scope=SettingScope.GLOBAL,
        user=None,
        value="light",
    )

    snapshot_payload = [{
        "key": "ui.theme.snapshot_admin",
        "type": "str",
        "title": "Theme",
        "description": "",
        "default": [],
        "choices": [],
        "required": False,
        "enabled": True,
        "editable": True,
        "frontend": False,
        "category": {"code": "ui", "title": "UI"},
        "has_global_value": True,
        "global_value": "dark",
    }]
    snapshot = type("SnapshotObj", (), {"pk": 1, "payload": snapshot_payload})()
    admin_obj = SettingsSnapshotAdmin(SettingsSnapshot, AdminSite())

    rendered = str(admin_obj.comparison_report(snapshot))

    assert "&quot;snapshot&quot;" in rendered
    assert "&quot;current&quot;" in rendered
    assert "&quot;only_in_snapshot&quot;" in rendered
    assert "&quot;only_in_current&quot;" in rendered
    assert "dark" in rendered
    assert "light" in rendered



from __future__ import annotations

import json

from django import forms
from django.contrib import admin, messages
from django.db.models import Count, QuerySet
from django.utils.html import escape, format_html

from .api import _ck  # внутренний хелпер для кэш-ключей
from .models import (
    SettingCategory,
    SettingDefinition,
    SettingScope,
    SettingsSnapshot,
    SettingType,
    SettingValue,
)
from .services import (
    build_global_settings_snapshot_payload,
    restore_global_settings_snapshot,
)
from .validators import validate_value


# ---------------------------
# Forms с типовой валидацией
# ---------------------------

class SettingDefinitionAdminForm(forms.ModelForm):
    class Meta:
        model = SettingDefinition
        fields = "__all__"

    def clean(self):
        cleaned = super().clean()
        defn: SettingDefinition = self.instance
        # при создании defn ещё нет pk, читаем из cleaned
        defn.type = cleaned.get("type", defn.type)
        defn.choices = cleaned.get("choices", defn.choices)
        defn.default = cleaned.get("default", defn.default)

        # простая проверка структуры choices для CHOICE
        if defn.type == SettingType.CHOICE and defn.choices:
            bad = [c for c in defn.choices if not isinstance(c, dict) or "value" not in c]
            if bad:
                raise forms.ValidationError(
                    'Для типа CHOICE поле \'choices\' должно быть списком объектов вида '
                    '[{"value": ..., "label": "..."}, ...].'
                )

        # валидируем default согласно типу
        try:
            if defn.default is not None:
                validate_value(defn, defn.default)
        except Exception as e:
            raise forms.ValidationError(f'Некорректное значение по умолчанию (default): {e}')

        return cleaned


class SettingValueAdminForm(forms.ModelForm):
    class Meta:
        model = SettingValue
        fields = "__all__"

    def clean(self):
        cleaned = super().clean()
        defn: SettingDefinition | None = cleaned.get("definition") or getattr(self.instance, "definition", None)
        value = cleaned.get("value", None)
        if defn is None:
            return cleaned
        try:
            # приведёт типы и проверит CHOICE/CRON/RRULE/JSON и т.п.
            validate_value(defn, value)
        except Exception as e:
            raise forms.ValidationError(f"Некорректное значение: {e}")
        return cleaned


class SettingsSnapshotAdminForm(forms.ModelForm):
    class Meta:
        model = SettingsSnapshot
        fields = "__all__"


# ---------------------------
# Inlines
# ---------------------------

class SettingValueInline(admin.TabularInline):
    model = SettingValue
    form = SettingValueAdminForm
    extra = 0
    raw_id_fields = ("user",)
    fields = ("scope", "user", "value")
    ordering = ("scope", "user")


# ---------------------------
# Admin actions
# ---------------------------

@admin.action(description="Очистить кэш для выбранных настроек")
def clear_cache_for_definitions(modeladmin, request, queryset: QuerySet[SettingDefinition]):
    from django.core.cache import cache
    from .models import SettingScope, SettingValue

    cleared = 0
    for d in queryset:
        # глобальный
        if cache.delete(_ck(d.key, None)):
            cleared += 1
        # пользовательские
        user_ids = SettingValue.objects.filter(
            definition=d, scope=SettingScope.USER
        ).values_list("user_id", flat=True)
        for uid in user_ids:
            if cache.delete(_ck(d.key, uid)):
                cleared += 1

    modeladmin.message_user(
        request,
        f"Инвалидировано кэш-ключей: {cleared}",
        level=messages.SUCCESS,
    )


@admin.action(description="Сохранить snapshot (снимок) всех глобальных настроек")
def create_settings_snapshot(modeladmin, request, queryset: QuerySet[SettingDefinition]):
    payload = build_global_settings_snapshot_payload()
    snapshot = SettingsSnapshot.objects.create(
        payload=payload,
        comment=f"Снимок через action, выбранных настроек: {queryset.count()}",
    )
    modeladmin.message_user(
        request,
        f"Создан снимок #{snapshot.pk}. В снимке: {len(payload)} настроек.",
        level=messages.SUCCESS,
    )


@admin.action(description="Восстановить настройки из snapshot (лишние настройки будут удалены)")
def restore_settings_snapshot(modeladmin, request, queryset: QuerySet[SettingsSnapshot]):
    snapshot = queryset.order_by("-created_at", "-id").first()
    if snapshot is None:
        modeladmin.message_user(request, "Не выбран снимок для восстановления.", level=messages.WARNING)
        return

    result = restore_global_settings_snapshot(snapshot.payload)
    modeladmin.message_user(
        request,
        (
            f"Восстановлен снимок #{snapshot.pk}. "
            f"Настроек создано: {result.created_definitions}, "
            f"обновлено: {result.updated_definitions}, "
            f"категорий создано: {result.created_categories}, "
            f"категорий обновлено: {result.updated_categories}, "
            f"глобальных значений обновлено: {result.updated_global_values}, "
            f"удалено лишних настроек: {result.deleted_definitions}."
        ),
        level=messages.SUCCESS,
    )


# ---------------------------
# ModelAdmins
# ---------------------------

@admin.register(SettingCategory)
class SettingCategoryAdmin(admin.ModelAdmin):
    list_display = ("code", "title")
    search_fields = ("code", "title")
    prepopulated_fields = {"code": ("title",)}
    ordering = ("code",)


@admin.register(SettingDefinition)
class SettingDefinitionAdmin(admin.ModelAdmin):
    form = SettingDefinitionAdminForm
    list_display = (
        "key",
        "title",
        "category",
        "type",
        "required",
        "enabled",
        "editable",
        "frontend",
        "values_count",
        "short_description",
    )
    list_filter = (
        "category",
        "type",
        "required",
        "enabled",
        "editable",
        "frontend",
        ("description", admin.EmptyFieldListFilter),
    )
    list_editable = (
        "required",
        "enabled",
        "editable",
        "frontend",
    )
    search_fields = ("key", "title", "description", "category__code", "category__title")
    ordering = ("category__code", "key")
    list_select_related = ("category",)
    list_display_links = ("key", "title")
    list_per_page = 50
    inlines = [SettingValueInline]
    actions = [clear_cache_for_definitions, create_settings_snapshot]
    autocomplete_fields = ()  # на будущее
    readonly_fields = ("values_count",)
    fieldsets = (
        ("Идентификация", {
            "fields": ("key", "title", "description", "category"),
        }),
        ("Тип и значения", {
            "fields": ("type", "default", "choices", "required"),
        }),
        ("Управление доступностью", {
            "fields": ("enabled", "editable", "frontend"),
        }),
        ("Статистика", {
            "fields": ("values_count",),
        }),
    )

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("category").annotate(_values_count=Count("values"))

    @admin.display(description="Значений", ordering="_values_count")
    def values_count(self, obj: SettingDefinition):
        return getattr(obj, "_values_count", 0)

    @admin.display(description="Описание")
    def short_description(self, obj: SettingDefinition):
        if not obj.description:
            return "—"
        text = obj.description.strip()
        return text if len(text) <= 120 else text[:117] + "…"


@admin.register(SettingValue)
class SettingValueAdmin(admin.ModelAdmin):
    form = SettingValueAdminForm
    list_display = ("definition_key", "scope", "user", "short_value")
    list_filter = ("scope", "definition__category", "definition__type")
    search_fields = ("definition__key", "definition__title", "user__username", "user__email")
    raw_id_fields = ("user", "definition")
    ordering = ("definition__key", "scope", "user")

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("definition", "definition__category", "user")

    @admin.display(description="definition", ordering="definition__key")
    def definition_key(self, obj: SettingValue):
        return obj.definition.key

    @admin.display(description="value")
    def short_value(self, obj: SettingValue):
        v = obj.value
        s = str(v)
        return s if len(s) <= 120 else s[:117] + "…"


@admin.register(SettingsSnapshot)
class SettingsSnapshotAdmin(admin.ModelAdmin):
    form = SettingsSnapshotAdminForm
    list_display = ("id", "created_at", "comment", "settings_count")
    search_fields = ("comment",)
    readonly_fields = ("created_at", "settings_count", "restore_mechanics", "comparison_report", "pretty_payload")
    fields = ("created_at", "comment", "settings_count", "restore_mechanics", "comparison_report", "pretty_payload", "payload")
    actions = [restore_settings_snapshot]
    ordering = ("-created_at", "-id")

    @admin.display(description="Кол-во настроек")
    def settings_count(self, obj: SettingsSnapshot):
        return len(obj.payload or [])

    @admin.display(description="Содержимое снимка")
    def pretty_payload(self, obj: SettingsSnapshot):
        return json.dumps(obj.payload, ensure_ascii=False, indent=2)

    @staticmethod
    def _build_snapshot_diff(
        snapshot_payload: list[dict],
        current_payload: list[dict],
    ) -> dict:
        snapshot_by_key = {item["key"]: item for item in snapshot_payload if item.get("key")}
        current_by_key = {item["key"]: item for item in current_payload if item.get("key")}

        snapshot_keys = set(snapshot_by_key)
        current_keys = set(current_by_key)

        only_in_snapshot = sorted(snapshot_keys - current_keys)
        only_in_current = sorted(current_keys - snapshot_keys)
        shared_keys = sorted(snapshot_keys & current_keys)

        changed: list[dict] = []
        unchanged_keys: list[str] = []

        for key in shared_keys:
            snapshot_item = snapshot_by_key[key]
            current_item = current_by_key[key]
            different_fields = sorted(
                field_name
                for field_name in (set(snapshot_item.keys()) | set(current_item.keys()))
                if snapshot_item.get(field_name) != current_item.get(field_name)
            )
            if different_fields:
                changed.append(
                    {
                        "key": key,
                        "different_fields": different_fields,
                        "snapshot": snapshot_item,
                        "current": current_item,
                    }
                )
            else:
                unchanged_keys.append(key)

        return {
            "counts": {
                "snapshot_total": len(snapshot_payload),
                "current_total": len(current_payload),
                "only_in_snapshot": len(only_in_snapshot),
                "only_in_current": len(only_in_current),
                "changed": len(changed),
                "unchanged": len(unchanged_keys),
            },
            "only_in_snapshot": [
                {"key": key, "snapshot": snapshot_by_key[key]}
                for key in only_in_snapshot
            ],
            "only_in_current": [
                {"key": key, "current": current_by_key[key]}
                for key in only_in_current
            ],
            "changed": changed,
            "unchanged_keys": unchanged_keys,
        }


    @admin.display(description="Как работает snapshot")
    def restore_mechanics(self, obj: SettingsSnapshot):
        return (
            "Restore применяет данные из payload к текущему реестру: "
            "обновляет существующие настройки, создает отсутствующие, "
            "обновляет глобальные значения и удаляет настройки, которых нет в snapshot."
        )

    @admin.display(description="Сравнение с текущими настройками")
    def comparison_report(self, obj: SettingsSnapshot):
        if obj.pk is None:
            return "Сохраните снимок, чтобы увидеть сравнение."

        current_payload = build_global_settings_snapshot_payload()
        diff = self._build_snapshot_diff(obj.payload or [], current_payload)
        rendered = escape(json.dumps(diff, ensure_ascii=False, indent=2))
        return format_html("<pre>{}</pre>", rendered)



from django.urls import path
from .views import SettingListView, SettingDetailView, SettingFrontendView

app_name = 'django_confetti'

urlpatterns = [
    path('settings/', SettingListView.as_view(), name='settings-list'),
    path('settings/frontend/', SettingFrontendView.as_view(), name='settings-frontend'),
    path('settings/<str:key>/', SettingDetailView.as_view(), name='settings-detail'),
]



import json, datetime

from .models import SettingDefinition
from .models import SettingType


def validate_value(defn: 'SettingDefinition', value):
    t = defn.type
    if t == SettingType.BOOL:
        if not isinstance(value, bool):
            raise ValueError('Ожидается bool')
    elif t == SettingType.INT: value = int(value)
    elif t == SettingType.FLOAT: value = float(value)
    elif t == SettingType.STR: value = str(value)
    elif t == SettingType.CHOICE:
        allowed = [c['value'] for c in (defn.choices or [])]
        if value not in allowed:
            raise ValueError(f'{defn.key} не допустимое значение для CHOICE')
    elif t == SettingType.JSON:
        json.dumps(value)
    elif t == SettingType.DATETIME:
        datetime.datetime.fromisoformat(value)
    elif t == SettingType.DURATION:
        iv = int(value)
        if iv < 0: raise ValueError('Продолжительность меньше 0')

    return value



import pytest
from confetti.api import get, set_value
from confetti.models import SettingScope, SettingDefinition
from django.core.cache import cache
from icecream import ic

pytestmark = pytest.mark.django_db


def test_get_returns_default_when_no_values():
    """Проверяет получение настройки."""

    assert get('ui.theme') == 'light'


def test_global_then_user_override_precedence(user):
    """Проверяет изменение user настроек, и что не трогает global."""

    set_value('ui.theme', 'dark', scope=SettingScope.GLOBAL)
    assert get('ui.theme') == 'dark', f'Не сменилось глобальное значение {get("ui.theme")} != dark'

    set_value('ui.theme', 'light', user=user)
    assert get('ui.theme', user=user) == 'light', f'У user не сменилось значение {get("ui.theme", user=user)} != light'
    assert get('ui.theme') == 'dark', f'Глобальная настройка изменилась {get("ui.theme")} != dark'


def test_cache_invalidation_on_set(user):
    """Простое изменение значений. проверка сброса кэш."""

    set_value('feature.jobs', True, scope=SettingScope.GLOBAL)
    assert get('feature.jobs') is True

    # теперь изменим значение → должно обновиться
    set_value('feature.jobs', False, scope=SettingScope.GLOBAL)
    assert get('feature.jobs') is False

def test_edit_not_editable(user):
    """Проверяет что нельзя изменить не редактируемую настройку."""
    assert get('edit') == False

    dfn = ic(SettingDefinition.objects.get(key='edit'))
    assert dfn.editable is False

    set_value('edit', True)
    dfn = SettingDefinition.objects.get(key='edit')
    assert dfn.editable is False
    assert get('edit') == False



from django.core.cache import cache

from .models import SettingDefinition, SettingValue, SettingScope
from .validators import validate_value
from .conf import confetti_settings

CACHE_PREFIX = confetti_settings.CACHE_PREFIX
CACHE_PREFIX_ENABLED = 'is_enabled'

def _ck(def_key, user_id=None):
    return f'{CACHE_PREFIX}:{def_key}:{user_id or "global"}'

def _is_enabled_ck(def_key, user_id=None):
    return f'{CACHE_PREFIX_ENABLED}:{def_key}:{user_id}:enabled'

def get(def_key: str, user=None, default=None):
    """
    Приоритет: user override -> global -> definition.default -> default (параметр).
    Результат кэшируется.
    """
    uid = getattr(user, 'id', None)
    for key in (_ck(def_key, uid), _ck(def_key, None)):
        cached = cache.get(key)
        if cached is not None and (uid or key.endswith(':global')):
            return cached
    try:
        defn = SettingDefinition.objects.get(key=def_key, enabled=True)
    except SettingDefinition.DoesNotExist:
        return default

    # user override
    if uid:
        sv = SettingValue.objects.filter(definition=defn, scope=SettingScope.USER, user_id=uid).first()

        if sv and sv.value is not None:
            cache.set(_ck(def_key, uid), sv.value)
            return sv.value

    # global
    gv = SettingValue.objects.filter(definition=defn, scope=SettingScope.GLOBAL, user__isnull=True).first()
    if gv and gv.value is not None:
        cache.set(_ck(def_key, None), gv.value)
        return gv.value
    return defn.default if defn.default is not None else default

def set_value(def_key: str, value, user=None, scope=None):
    """
    Устанавливает значение с валидацией по типу.
    scope: None => USER если user передан, иначе GLOBAL.
    """
    defn = SettingDefinition.objects.get(key=def_key)

    value = validate_value(defn, value)
    if scope is None:
        scope = SettingScope.USER if user else SettingScope.GLOBAL

    sv, _ = SettingValue.objects.get_or_create(
        definition=defn, scope=scope, user=user if scope == SettingScope.USER else None
    )
    if not defn.editable:
        value = defn.default
    sv.value = value
    sv.save()
    cache.set(_ck(def_key, sv.user_id if sv.scope == SettingScope.USER else None), value)
    return sv

def is_enabled(flag_key: str, user=None, default=False) -> bool:
    """
    Проверяет, включена ли настройка (definition.enabled == True)
    и возвращает её логическое значение.
    Приоритет: user override -> global -> definition.default -> default (параметр).
    Результат кэшируется отдельно.
    """
    uid = getattr(user, 'id', None)
    cache_key = _is_enabled_ck(flag_key, uid)
    cached = cache.get(cache_key)
    if cached is not None:
        return cached

    try:
        defn = SettingDefinition.objects.get(key=flag_key)
    except SettingDefinition.DoesNotExist:
        cache.set(cache_key, default)
        return default

    # Если настройка выключена на уровне definition.enabled — она выключена для всех
    if not defn.enabled:
        cache.set(cache_key, False)
        return False

    # user override
    if uid:
        sv = SettingValue.objects.filter(
            definition=defn, scope=SettingScope.USER, user_id=uid
        ).first()
        if sv and sv.value is not None:
            result = bool(sv.value)
            cache.set(cache_key, result)
            return result

    # global
    gv = SettingValue.objects.filter(
        definition=defn, scope=SettingScope.GLOBAL, user__isnull=True
    ).first()
    if gv and gv.value is not None:
        result = bool(gv.value)
        cache.set(cache_key, result)
        return result

    # иначе используем default
    result = bool(defn.default) if defn.default is not None else bool(default)
    cache.set(cache_key, result)
    return result



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from django.core.cache import cache
from django.db import transaction

from .api import _ck, _is_enabled_ck
from .models import (
    SettingCategory,
    SettingDefinition,
    SettingScope,
    SettingValue,
)


SETTING_DEFINITION_SNAPSHOT_FIELDS = (
    "key",
    "type",
    "title",
    "description",
    "default",
    "choices",
    "required",
    "enabled",
    "editable",
    "frontend",
)


@dataclass
class RestoreResult:
    created_definitions: int = 0
    updated_definitions: int = 0
    created_categories: int = 0
    updated_categories: int = 0
    updated_global_values: int = 0
    deleted_definitions: int = 0


def build_global_settings_snapshot_payload() -> list[dict[str, Any]]:
    definitions = (
        SettingDefinition.objects
        .select_related("category")
        .prefetch_related("values")
        .order_by("key")
    )

    payload: list[dict[str, Any]] = []
    for definition in definitions:
        global_value_obj = next(
            (
                value
                for value in definition.values.all()
                if value.scope == SettingScope.GLOBAL and value.user_id is None
            ),
            None,
        )

        item = {
            field_name: getattr(definition, field_name)
            for field_name in SETTING_DEFINITION_SNAPSHOT_FIELDS
        }
        item["category"] = None
        if definition.category:
            item["category"] = {
                "code": definition.category.code,
                "title": definition.category.title,
            }
        item["has_global_value"] = global_value_obj is not None
        item["global_value"] = global_value_obj.value if global_value_obj else None
        payload.append(item)

    return payload


@transaction.atomic
def restore_global_settings_snapshot(payload: list[dict[str, Any]]) -> RestoreResult:
    result = RestoreResult()

    snapshot_keys = {
        item.get("key")
        for item in payload
        if item.get("key")
    }
    stale_definitions = SettingDefinition.objects.exclude(key__in=snapshot_keys)

    stale_keys = list(stale_definitions.values_list("key", flat=True))
    result.deleted_definitions = stale_definitions.count()
    stale_definitions.delete()

    for stale_key in stale_keys:
        cache.delete(_ck(stale_key, None))
        cache.delete(_is_enabled_ck(stale_key, None))

    for item in payload:
        category_obj = None
        category_data = item.get("category")
        if category_data and category_data.get("code"):
            category_defaults = {"title": category_data.get("title", category_data["code"])}
            category_obj, created_category = SettingCategory.objects.get_or_create(
                code=category_data["code"],
                defaults=category_defaults,
            )
            if created_category:
                result.created_categories += 1
            elif category_data.get("title") is not None and category_obj.title != category_data["title"]:
                category_obj.title = category_data["title"]
                category_obj.save(update_fields=["title"])
                result.updated_categories += 1

        defaults = {
            field_name: item.get(field_name)
            for field_name in SETTING_DEFINITION_SNAPSHOT_FIELDS
            if field_name != "key"
        }
        defaults["category"] = category_obj

        definition, created_definition = SettingDefinition.objects.get_or_create(
            key=item["key"],
            defaults=defaults,
        )

        if created_definition:
            result.created_definitions += 1
        else:
            changed_fields: list[str] = []
            for field_name, field_value in defaults.items():
                if getattr(definition, field_name) != field_value:
                    setattr(definition, field_name, field_value)
                    changed_fields.append(field_name)

            if changed_fields:
                definition.save(update_fields=changed_fields)
                result.updated_definitions += 1

        if item.get("has_global_value"):
            SettingValue.objects.update_or_create(
                definition=definition,
                scope=SettingScope.GLOBAL,
                user=None,
                defaults={"value": item.get("global_value")},
            )
            result.updated_global_values += 1

        cache.delete(_ck(definition.key, None))
        cache.delete(_is_enabled_ck(definition.key, None))

    return result



from __future__ import annotations

try:
    from drf_yasg.utils import swagger_auto_schema as _swagger_auto_schema
    from drf_yasg import openapi as _openapi
    DRF_YASG_AVAILABLE = True
except Exception:
    DRF_YASG_AVAILABLE = False
    _swagger_auto_schema = None
    _openapi = None


def swagger_auto_schema(*arfs, **kwargs):
    """Ноу-оп декоратор, если drf-yasg отсутствует."""
    if not DRF_YASG_AVAILABLE:
        def _decorator(view):
            return view
        return _decorator
    return _swagger_auto_schema(*arfs, **kwargs)

class openapi:
    if DRF_YASG_AVAILABLE:
        Schema = _openapi.Schema
        Response = _openapi.Response
        Parameter = _openapi.Parameter
        TYPE_OBJECT = _openapi.TYPE_OBJECT
        TYPE_ARRAY = _openapi.TYPE_ARRAY
        TYPE_STRING = _openapi.TYPE_STRING
        IN_PATH = _openapi.IN_PATH
    else:
        TYPE_OBJECT = 'object'
        TYPE_ARRAY = 'array'
        TYPE_STRING = 'string'
        IN_PATH = 'path'

        class Schema:
            def __init__(self, **kwargs): pass

        class Response:
            def __init__(self, *args, **kwargs): pass

        class Parameter:
            def __init__(self, *args, **kwargs): pass



import pytest

from confetti.models import SettingCategory, SettingDefinition, SettingType


@pytest.mark.django_db
def test_bool_definition_syncs_default_on_create_from_enabled():
    category, _ = SettingCategory.objects.get_or_create(code="flags", defaults={"title": "Flags"})

    definition = SettingDefinition.objects.create(
        key="feature.create-sync",
        category=category,
        type=SettingType.BOOL,
        title="Create sync",
        enabled=False,
        default=True,
    )

    assert definition.default is False


@pytest.mark.django_db
def test_bool_definition_syncs_default_when_enabled_toggles():
    category, _ = SettingCategory.objects.get_or_create(code="flags", defaults={"title": "Flags"})
    definition = SettingDefinition.objects.create(
        key="feature.toggle-sync",
        category=category,
        type=SettingType.BOOL,
        title="Toggle sync",
        enabled=True,
        default=True,
    )
    assert definition.default is True

    definition.enabled = False
    definition.save()
    definition.refresh_from_db()

    assert definition.default is False

    definition.enabled = True
    definition.save()
    definition.refresh_from_db()

    assert definition.default is True
